==> postgresql-1.0.2.tgz: package/src/index.ts <==
import {
  RegionType,
  ApiCredentials,
  removeUndefined,
  waitResponse,
} from '@tencent-serverless-framework/common';
import { Client } from 'tencentcloud-sdk-nodejs/tencentcloud/services/postgres/v20170312/postgres_client';
import {
  PostgresqlDeployInputs,
  PostgresqlDeployOutputs,
  PostgresqlRemoveInputs,
} from './interface';
import {
  CreateServerlessDBInstanceRequest,
  ServerlessDBAccount,
  ServerlessDBInstance,
  ServerlessDBInstanceNetInfo,
} from 'tencentcloud-sdk-nodejs/tencentcloud/services/postgres/v20170312/postgres_models';

// timeout 5 minutes
const TIMEOUT = 5 * 60 * 1000;

/** Postgresql （Postgresql 数据库）for serverless */
export class Postgresql {
  credentials: ApiCredentials;
  region: RegionType;
  client: Client;

  /**
   * Postgresql for serverless 构造函数，返回 client
   *
   * @param credentials 认证信息
   * @param region 区域
   */
  constructor(credentials: ApiCredentials = {}, region: RegionType) {
    this.region = region || 'ap-guangzhou';
    this.credentials = credentials;
    this.client = new Client({
      credential: removeUndefined({
        secretId: this.credentials.SecretId!,
        secretKey: this.credentials.SecretKey!,
        soken: this.credentials.Token,
      }),
      region: this.region,
      profile: {},
    });
  }

  /**
   * 部署 postgresql 实例
   *
   * @param inputs  输入
   * @returns 部署结果
   */
  async deploy(inputs: PostgresqlDeployInputs = {}): Promise<PostgresqlDeployOutputs> {
    const {
      region,
      zone,
      projectId,
      dBInstanceName,
      dBVersion,
      dBCharset,
      extranetAccess,
      vpcConfig,
    } = inputs;

    const outputs: PostgresqlDeployOutputs = {
      region: region,
      zone: zone,
      vpcConfig: vpcConfig,
      dBInstanceName: dBInstanceName,
    };

    let dbDetail = await this.getDbInstanceDetail(dBInstanceName!);

    // FIXME: 不同可用区下若存在同名数据库，可能导致重复创建
    /** 数据库实例名，所在可用区相同，不再重新创建数据库 */
    if (dbDetail && dbDetail.DBInstanceName && dbDetail.Zone === zone) {
      const publicAccess = this.getDbExtranetAccess(dbDetail.DBInstanceNetInfo!);
      // exist and public access config different, update db instance
      if (publicAccess !== extranetAccess) {
        console.log(`DB instance ${dBInstanceName} existed, updating`);
        // do not throw error when open public access
        try {
          dbDetail = await this.toggleDbInstanceAccess(dBInstanceName!, extranetAccess!);
        } catch (e) {
          console.warn(`Toggle DB Instane access failed, ${e.message}, ${e.reqId}`);
        }
      } else {
        console.log(`DB instance ${dBInstanceName} existed.`);
      }
    } else {
      // not exist, create
      const postgresInputs = {
        Zone: zone!,
        ProjectId: projectId!,
        DBInstanceName: dBInstanceName!,
        DBVersion: dBVersion!,
        DBCharset: dBCharset!,
        VpcId: vpcConfig?.vpcId!,
        SubnetId: vpcConfig?.subnetId!,
      };

      dbDetail = await this.createDbInstance(postgresInputs);
      if (extranetAccess) {
        dbDetail = await this.toggleDbInstanceAccess(dBInstanceName!, extranetAccess);
      }
    }
    outputs.dBInstanceId = dbDetail?.DBInstanceId;

    const { DBInstanceNetInfo, DBAccountSet, DBDatabaseList } = dbDetail!;
    // FIXME: 有多个账号或多个数据库时，只能取到第一个？
    const [accountInfo] = DBAccountSet!;
    const [dbName] = DBDatabaseList!;

    let internetInfo: ServerlessDBInstanceNetInfo;
    let extranetInfo: ServerlessDBInstanceNetInfo;

    DBInstanceNetInfo?.forEach((item) => {
      if (item.NetType === 'private') {
        internetInfo = item;
      }
      if (item.NetType === 'public') {
        extranetInfo = item;
      }
    });
    if (vpcConfig?.vpcId) {
      outputs.private = this.formatPgUrl(internetInfo!, accountInfo, dbName);
    }
    if (extranetAccess && extranetInfo!) {
      outputs.public = this.formatPgUrl(extranetInfo, accountInfo, dbName);
    }

    return outputs;
  }

  /**
   * 移除 postgresql 实例
   *
   * @param inputs 移除 postgresql 实例输入
   * @returns 移除结果，成功返回 {}
   */
  async remove(inputs: PostgresqlRemoveInputs = {}) {
    const { dBInstanceName } = inputs;

    const dbDetail = await this.getDbInstanceDetail(dBInstanceName!);
    if (dbDetail && dbDetail.DBInstanceName) {
      // need circle for deleting, after host status is 6, then we can delete it
      await this.deleteDbInstance(dBInstanceName!);
    }
    return {};
  }

  /**
   * 通过数据库名称获取实例
   *
   * @param dBInstanceName 数据库名称
   * @returns 数据库详情
   */
  async getDbInstanceDetail(dBInstanceName: string) {
    // get instance detail
    try {
      const res = await this.client.DescribeServerlessDBInstances({
        Filter: [
          {
            Name: 'db-instance-name',
            Values: [dBInstanceName],
          },
        ],
      });
      if (res.DBInstanceSet) {
        const {
          DBInstanceSet: [dbDetail],
        } = res;
        return dbDetail;
      }
    } catch (e) {
      console.warn(e);
    }
    return undefined;
  }

  /**
   * 获取数据库是否允许外部访问
   *
   * @param netInfos 数据库网络信息
   * @returns 数据库是否允许外网访问
   */
  getDbExtranetAccess(netInfos: ServerlessDBInstanceNetInfo[]) {
    let result = false;
    netInfos.forEach((item) => {
      if (item.NetType === 'public') {
        result = item.Status === 'opened';
      }
    });
    return result;
  }

  /*
 INSTANCE_STATUS_APPLYING:    "applying",    申请中
 INSTANCE_STATUS_INIT:        "init",        待初始化
 INSTANCE_STATUS_INITING:     "initing",     初始化中
 INSTANCE_STATUS_OK:          "running",     运行中
 INSTANCE_STATUS_LIMITED:     "limited run", 受限运行
 INSTANCE_STATUS_ISOLATED:    "isolated",    已隔离
 INSTANCE_STATUS_RECYCLING:   "recycling",   回收中
 INSTANCE_STATUS_RECYCLED:    "recycled",    已回收
 INSTANCE_STATUS_JOB_RUNNING: "job running", 任务执行中
 INSTANCE_STATUS_OFFLINE:     "offline",     下线
 INSTANCE_STATUS_MIGRATE:     "migrating",   迁移中
 INSTANCE_STATUS_EXPANDING:   "expanding",   扩容中
 INSTANCE_STATUS_READONLY:    "readonly",    只读
 INSTANCE_STATUS_RESTARTING:  "restarting",  重启中
 */

  /**
   * 设置数据库外网访问
   *
   * @param dBInstanceName 数据库实例名称
   * @param extranetAccess 设置是否有外网访问权限
   * @returns 返回数据库详情
   */
  async toggleDbInstanceAccess(dBInstanceName: string, extranetAccess: boolean) {
    const res = await this.getDbInstanceDetail(dBInstanceName);

    const { DBInstanceId } = res!;
    if (extranetAccess) {
      console.log(`Start open db extranet access...`);
      // FIXME: pass dbinstanceName cause Error: 参数DBInstanceId|DBInstanceIdSet有误
      await this.client.OpenServerlessDBExtranetAccess({
        DBInstanceId,
      });
      const detail: Promise<ServerlessDBInstance | undefined> = await waitResponse({
        callback: async () => this.getDbInstanceDetail(dBInstanceName),
        targetResponse: 'running',
        targetProp: 'DBInstanceStatus',
        timeout: TIMEOUT,
      });
      console.log(`Open db extranet access success`);
      return detail;
    }

    console.log(`Start close db extranet access`);
    await this.client.CloseServerlessDBExtranetAccess({
      DBInstanceId,
    });
    const detail: Promise<ServerlessDBInstance | undefined> = await waitResponse({
      callback: async () => this.getDbInstanceDetail(dBInstanceName),
      targetResponse: 'running',
      targetProp: 'DBInstanceStatus',
      timeout: TIMEOUT,
    });
    console.log(`Close db extranet access success`);
    return detail;
  }

  /**
   * 创建数据库实例
   *
   * @param postgresInputs 创建实例输入参数
   * @returns 返回数据库详情
   */
  async createDbInstance(postgresInputs: CreateServerlessDBInstanceRequest) {
    console.log(`Start create DB instance ${postgresInputs.DBInstanceName}`);
    const { DBInstanceId } = await this.client.CreateServerlessDBInstance(postgresInputs);
    console.log(`Creating DB instance ID: ${DBInstanceId}`);

    const detail: Promise<ServerlessDBInstance | undefined> = await waitResponse({
      callback: async () => this.getDbInstanceDetail(postgresInputs.DBInstanceName),
      targetResponse: 'running',
      targetProp: 'DBInstanceStatus',
      timeout: TIMEOUT,
    });
    console.log(`Created DB instance name ${postgresInputs.DBInstanceName} successfully`);
    return detail;
  }

  /**
   * 删除数据库实例
   *
   * @param dBInstanceName 数据库实例名称
   * @returns 返回删除后的实例信息，如成功为 undefined
   */
  async deleteDbInstance(dBInstanceName: string) {
    console.log(`Start removing postgres instance ${dBInstanceName}`);
    const res = await this.getDbInstanceDetail(dBInstanceName);

    // FIXME: pass dbinstanceName cause Error: 参数DBInstanceId|DBInstanceIdSet有误
    const { DBInstanceId } = res!;

    await this.client.DeleteServerlessDBInstance({
      DBInstanceId,
    });
    const detail: Promise<ServerlessDBInstance | undefined> = await waitResponse({
      callback: async () => this.getDbInstanceDetail(dBInstanceName),
      targetResponse: undefined,
      timeout: TIMEOUT,
    });
    console.log(`Removed postgres instance ${dBInstanceName} successfully`);
    return detail;
  }

  /**
   * 格式化数据库连接串
   *
   * @param netInfo 数据库网络信息
   * @param accountInfo 数据库账户信息
   * @param dbName 数据库名称
   * @returns 返回数据库连接串 object
   */
  formatPgUrl(
    netInfo: ServerlessDBInstanceNetInfo,
    accountInfo: ServerlessDBAccount,
    dbName: string,
  ) {
    return {
      connectionString: `postgresql://${accountInfo.DBUser}:${encodeURIComponent(
        accountInfo.DBPassword,
      )}@${netInfo.Address || netInfo.Ip}:${netInfo.Port}/${dbName}`,
      host: netInfo.Address || netInfo.Ip,
      port: netInfo.Port,
      user: accountInfo.DBUser,
      password: accountInfo.DBPassword,
      dbname: dbName,
    };
  }
}

==> postgresql-1.0.4.tgz: package/src/index.ts <==
import {
  RegionType,
  ApiCredentials,
  removeUndefined,
  waitResponse,
} from '@tencent-serverless-framework/common';
import { Client } from 'tencentcloud-sdk-nodejs/tencentcloud/services/postgres/v20170312/postgres_client';
import {
  PostgresqlDeployInputs,
  PostgresqlDeployOutputs,
  PostgresqlRemoveInputs,
} from './interface';
import {
  CreateServerlessDBInstanceRequest,
  ServerlessDBAccount,
  ServerlessDBInstance,
  ServerlessDBInstanceNetInfo,
} from 'tencentcloud-sdk-nodejs/tencentcloud/services/postgres/v20170312/postgres_models';

// timeout 5 minutes
const TIMEOUT = 5 * 60 * 1000;

/** Postgresql （Postgresql 数据库）for serverless component */
export class Postgresql {
  /** 地域 */
  region: RegionType;
  /** 鉴权信息 */
  credentials: ApiCredentials;
  /** sdk 实例 */
  client: Client;

  /**
   * Postgresql for serverless component 构造函数，返回 client
   *
   * @param credentials 认证信息
   * @param region 区域
   */
  constructor(credentials: ApiCredentials = {}, region: RegionType) {
    this.region = region || 'ap-guangzhou';
    this.credentials = credentials;
    this.client = new Client({
      credential: removeUndefined({
        secretId: this.credentials.SecretId!,
        secretKey: this.credentials.SecretKey!,
        soken: this.credentials.Token,
      }),
      region: this.region,
      profile: {},
    });
  }

  /**
   * 部署 postgresql 实例
   *
   * @param inputs  输入
   * @returns 部署结果
   */
  async deploy(inputs: PostgresqlDeployInputs = {}): Promise<PostgresqlDeployOutputs> {
    const {
      region,
      zone,
      projectId,
      dBInstanceName,
      dBVersion,
      dBCharset,
      extranetAccess,
      vpcConfig,
    } = inputs;

    const outputs: PostgresqlDeployOutputs = {
      region: region,
      zone: zone,
      vpcConfig: vpcConfig,
      dBInstanceName: dBInstanceName,
    };

    let dbDetail = await this.getDbInstanceDetail(dBInstanceName!);

    // FIXME: 不同可用区下若存在同名数据库，可能导致重复创建
    /** 数据库实例名，所在可用区相同，不再重新创建数据库 */
    if (dbDetail && dbDetail.DBInstanceName && dbDetail.Zone === zone) {
      const publicAccess = this.getDbExtranetAccess(dbDetail.DBInstanceNetInfo!);
      // exist and public access config different, update db instance
      if (publicAccess !== extranetAccess) {
        console.log(`DB instance ${dBInstanceName} existed, updating`);
        // do not throw error when open public access
        try {
          dbDetail = await this.toggleDbInstanceAccess(dBInstanceName!, extranetAccess!);
        } catch (e) {
          console.warn(`Toggle DB Instane access failed, ${e.message}, ${e.reqId}`);
        }
      } else {
        console.log(`DB instance ${dBInstanceName} existed.`);
      }
    } else {
      // not exist, create
      const postgresInputs = {
        Zone: zone!,
        ProjectId: projectId!,
        DBInstanceName: dBInstanceName!,
        DBVersion: dBVersion!,
        DBCharset: dBCharset!,
        VpcId: vpcConfig?.vpcId!,
        SubnetId: vpcConfig?.subnetId!,
      };

      dbDetail = await this.createDbInstance(postgresInputs);
      if (extranetAccess) {
        dbDetail = await this.toggleDbInstanceAccess(dBInstanceName!, extranetAccess);
      }
    }
    outputs.dBInstanceId = dbDetail?.DBInstanceId;

    const { DBInstanceNetInfo, DBAccountSet, DBDatabaseList } = dbDetail!;
    // FIXME: 有多个账号或多个数据库时，只能取到第一个？
    const [accountInfo] = DBAccountSet!;
    const [dbName] = DBDatabaseList!;

    let internetInfo: ServerlessDBInstanceNetInfo;
    let extranetInfo: ServerlessDBInstanceNetInfo;

    DBInstanceNetInfo?.forEach((item) => {
      if (item.NetType === 'private') {
        internetInfo = item;
      }
      if (item.NetType === 'public') {
        extranetInfo = item;
      }
    });
    if (vpcConfig?.vpcId) {
      outputs.private = this.formatPgUrl(internetInfo!, accountInfo, dbName);
    }
    if (extranetAccess && extranetInfo!) {
      outputs.public = this.formatPgUrl(extranetInfo, accountInfo, dbName);
    }

    return outputs;
  }

  /**
   * 移除 postgresql 实例
   *
   * @param inputs 移除 postgresql 实例输入
   * @returns 移除结果，成功返回 {}
   */
  async remove(inputs: PostgresqlRemoveInputs = {}) {
    const { dBInstanceName } = inputs;

    const dbDetail = await this.getDbInstanceDetail(dBInstanceName!);
    if (dbDetail && dbDetail.DBInstanceName) {
      // need circle for deleting, after host status is 6, then we can delete it
      await this.deleteDbInstance(dBInstanceName!);
    }
    return {};
  }

  /**
   * 通过数据库名称获取实例
   *
   * @param dBInstanceName 数据库名称
   * @returns 数据库详情
   */
  async getDbInstanceDetail(dBInstanceName: string) {
    // get instance detail
    try {
      const res = await this.client.DescribeServerlessDBInstances({
        Filter: [
          {
            Name: 'db-instance-name',
            Values: [dBInstanceName],
          },
        ],
      });
      if (res.DBInstanceSet) {
        const {
          DBInstanceSet: [dbDetail],
        } = res;
        return dbDetail;
      }
    } catch (e) {
      console.warn(e);
    }
    return undefined;
  }

  /**
   * 获取数据库是否允许外部访问
   *
   * @param netInfos 数据库网络信息
   * @returns 数据库是否允许外网访问
   */
  getDbExtranetAccess(netInfos: ServerlessDBInstanceNetInfo[]) {
    let result = false;
    netInfos.forEach((item) => {
      if (item.NetType === 'public') {
        result = item.Status === 'opened';
      }
    });
    return result;
  }

  /*
 INSTANCE_STATUS_APPLYING:    "applying",    申请中
 INSTANCE_STATUS_INIT:        "init",        待初始化
 INSTANCE_STATUS_INITING:     "initing",     初始化中
 INSTANCE_STATUS_OK:          "running",     运行中
 INSTANCE_STATUS_LIMITED:     "limited run", 受限运行
 INSTANCE_STATUS_ISOLATED:    "isolated",    已隔离
 INSTANCE_STATUS_RECYCLING:   "recycling",   回收中
 INSTANCE_STATUS_RECYCLED:    "recycled",    已回收
 INSTANCE_STATUS_JOB_RUNNING: "job running", 任务执行中
 INSTANCE_STATUS_OFFLINE:     "offline",     下线
 INSTANCE_STATUS_MIGRATE:     "migrating",   迁移中
 INSTANCE_STATUS_EXPANDING:   "expanding",   扩容中
 INSTANCE_STATUS_READONLY:    "readonly",    只读
 INSTANCE_STATUS_RESTARTING:  "restarting",  重启中
 */

  /**
   * 设置数据库外网访问
   *
   * @param dBInstanceName 数据库实例名称
   * @param extranetAccess 设置是否有外网访问权限
   * @returns 返回数据库详情
   */
  async toggleDbInstanceAccess(dBInstanceName: string, extranetAccess: boolean) {
    const res = await this.getDbInstanceDetail(dBInstanceName);

    const { DBInstanceId } = res!;
    if (extranetAccess) {
      console.log(`Start open db extranet access...`);
      // FIXME: pass dbinstanceName cause Error: 参数DBInstanceId|DBInstanceIdSet有误
      await this.client.OpenServerlessDBExtranetAccess({
        DBInstanceId,
      });
      const detail: Promise<ServerlessDBInstance | undefined> = await waitResponse({
        callback: async () => this.getDbInstanceDetail(dBInstanceName),
        targetResponse: 'running',
        targetProp: 'DBInstanceStatus',
        timeout: TIMEOUT,
      });
      console.log(`Open db extranet access success`);
      return detail;
    }

    console.log(`Start close db extranet access`);
    await this.client.CloseServerlessDBExtranetAccess({
      DBInstanceId,
    });
    const detail: Promise<ServerlessDBInstance | undefined> = await waitResponse({
      callback: async () => this.getDbInstanceDetail(dBInstanceName),
      targetResponse: 'running',
      targetProp: 'DBInstanceStatus',
      timeout: TIMEOUT,
    });
    console.log(`Close db extranet access success`);
    return detail;
  }

  /**
   * 创建数据库实例
   *
   * @param postgresInputs 创建实例输入参数
   * @returns 返回数据库详情
   */
  async createDbInstance(postgresInputs: CreateServerlessDBInstanceRequest) {
    console.log(`Start create DB instance ${postgresInputs.DBInstanceName}`);
    const { DBInstanceId } = await this.client.CreateServerlessDBInstance(postgresInputs);
    console.log(`Creating DB instance ID: ${DBInstanceId}`);

    const detail: Promise<ServerlessDBInstance | undefined> = await waitResponse({
      callback: async () => this.getDbInstanceDetail(postgresInputs.DBInstanceName),
      targetResponse: 'running',
      targetProp: 'DBInstanceStatus',
      timeout: TIMEOUT,
    });
    console.log(`Created DB instance name ${postgresInputs.DBInstanceName} successfully`);
    return detail;
  }

  /**
   * 删除数据库实例
   *
   * @param dBInstanceName 数据库实例名称
   * @returns 返回删除后的实例信息，如成功为 undefined
   */
  async deleteDbInstance(dBInstanceName: string) {
    console.log(`Start removing postgres instance ${dBInstanceName}`);
    const res = await this.getDbInstanceDetail(dBInstanceName);

    // FIXME: pass dbinstanceName cause Error: 参数DBInstanceId|DBInstanceIdSet有误
    const { DBInstanceId } = res!;

    await this.client.DeleteServerlessDBInstance({
      DBInstanceId,
    });
    const detail: Promise<ServerlessDBInstance | undefined> = await waitResponse({
      callback: async () => this.getDbInstanceDetail(dBInstanceName),
      targetResponse: undefined,
      timeout: TIMEOUT,
    });
    console.log(`Removed postgres instance ${dBInstanceName} successfully`);
    return detail;
  }

  /**
   * 格式化数据库连接串
   *
   * @param netInfo 数据库网络信息
   * @param accountInfo 数据库账户信息
   * @param dbName 数据库名称
   * @returns 返回数据库连接串 object
   */
  formatPgUrl(
    netInfo: ServerlessDBInstanceNetInfo,
    accountInfo: ServerlessDBAccount,
    dbName: string,
  ) {
    return {
      connectionString: `postgresql://${accountInfo.DBUser}:${encodeURIComponent(
        accountInfo.DBPassword,
      )}@${netInfo.Address || netInfo.Ip}:${netInfo.Port}/${dbName}`,
      host: netInfo.Address || netInfo.Ip,
      port: netInfo.Port,
      user: accountInfo.DBUser,
      password: accountInfo.DBPassword,
      dbname: dbName,
    };
  }
}
